"""AtcoderHelperConfigを永続化する層."""

import os
from typing import Protocol

import yaml

from atcoder_helper.models.atcoder_helper_config import AtCoderHelperConfig
from atcoder_helper.repositories.errors import ParseError
from atcoder_helper.repositories.errors import ReadError
from atcoder_helper.repositories.errors import WriteError
from atcoder_helper.services.util import get_atcoder_helper_config_filepath


class ConfigRepository(Protocol):
    """AtCoderHelperConfigを永続化するプロトコル."""

    def __init__(self, filename: str):
        """__init__."""

    def read(self) -> AtCoderHelperConfig:
        """読み込みを行う.

        Raises:
            ReadError: 読み込みに失敗した
            ParseError: パースに失敗した

        Returns:
            AtCoderHelperConfig: 読み込まれたAtcoderHelperConfig
        """

    def write(self, config: AtCoderHelperConfig) -> None:
        """書き込みを行う.

        Args:
            config (AtCoderHelperConfig): _description_

        Raises:
            WriteError: 書き込みに失敗した
        """


def get_default_config_repository() -> ConfigRepository:
    """ConfigRepositoryの標準実装を返す."""
    return ConfigRepositoryImpl(get_atcoder_helper_config_filepath())


class ConfigRepositoryImpl:
    """AtCoderHelperConfigを永続化する層."""

    _filename: str

    def __init__(self, filename: str) -> None:
        """__init__."""
        self._filename = filename

    def read(self) -> AtCoderHelperConfig:
        """読み込みを行う.

        Raises:
            ReadError: 読み込みに失敗した
            ParseError: パースに失敗した

        Returns:
            AtCoderHelperConfig: 読み込まれたAtcoderHelperConfig
        """
        try:
            with open(self._filename, "rt") as file:
                config_obj = yaml.safe_load(file)
        except OSError as e:
            raise ReadError(f"cannot read from {self._filename}.") from e

        try:
            language_dict = {
                language["name"]: language for language in config_obj["languages"]
            }
            return AtCoderHelperConfig.parse_obj(
                {
                    "languages": language_dict,
                    "default_language": config_obj["default_language"],
                }
            )
        except Exception as e:
            raise ParseError(
                f"{self._filename} can not read as AtCoderHelperConfig"
            ) from e

    def write(self, config: AtCoderHelperConfig) -> None:
        """書き込みを行う.

        Args:
            config (AtCoderHelperConfig): _description_

        Raises:
            WriteError: 書き込みに失敗した
        """
        os.makedirs(name=os.path.dirname(self._filename), exist_ok=True)

        try:
            with open(self._filename, "wt") as file:
                yaml.dump(config.dict(exclude_none=True), file)
        except OSError as e:
            raise WriteError("cannot write to {self._filename}.") from e



"""logged_in_session_repoのテスト."""


from typing import Type

import mock
import pytest
import requests
from pytest import MonkeyPatch

from atcoder_helper.repositories.atcoder_logged_in_session_repo import (
    AtCoderLoggedInSessionRepositoryImpl,
)
from atcoder_helper.repositories.errors import ConnectionError
from atcoder_helper.repositories.errors import LoginFailure
from atcoder_helper.repositories.errors import ParseError

test_read_parameter = {
    "OK": [
        mock.MagicMock(return_value=mock.MagicMock(text='<input value="foo"/>')),
        mock.MagicMock(return_value=mock.MagicMock(headers={"Location": "/home"})),
        None,
    ],
    "error(get失敗)": [
        mock.MagicMock(side_effect=Exception()),
        mock.MagicMock(),
        ConnectionError,
    ],
    "error(parser失敗)": [
        mock.MagicMock(return_value=mock.MagicMock(text="<input/>")),
        mock.MagicMock(return_value=mock.MagicMock(headers={"Location": "/home"})),
        ParseError,
    ],
    "error(POST失敗)": [
        mock.MagicMock(return_value=mock.MagicMock(text='<input value="foo"/>')),
        mock.MagicMock(side_effect=Exception),
        ConnectionError,
    ],
    "error(ログイン失敗)": [
        mock.MagicMock(return_value=mock.MagicMock(text='<input value="foo"/>')),
        mock.MagicMock(return_value=mock.MagicMock(headers={"Location": "/foo"})),
        LoginFailure,
    ],
}


@pytest.mark.parametrize(
    argnames=("get_mock", "post_mock", "exception"),
    argvalues=test_read_parameter.values(),
    ids=test_read_parameter.keys(),
)
def test_read(
    get_mock: mock.MagicMock,
    post_mock: mock.MagicMock,
    exception: Type[Exception],
    monkeypatch: MonkeyPatch,
) -> None:
    """readのテスト.

    sessionの検証は難しいのでやらない
    """
    username = "name"
    password = "password"
    sut = AtCoderLoggedInSessionRepositoryImpl()

    monkeypatch.setattr(requests.Session, "get", get_mock)
    monkeypatch.setattr(requests.Session, "post", post_mock)

    if exception:
        with pytest.raises(exception):
            sut.read(username, password)
    else:
        sut.read(username, password)



"""プログラムを実行して結果を取得する."""

import subprocess
from typing import List
from typing import Protocol

from atcoder_helper.models.atcoder_test_case import AtcoderTestCase
from atcoder_helper.models.atcoder_test_case import AtCoderTestResult
from atcoder_helper.models.atcoder_test_case import AtCoderTestStatus


class ProgramExecutor(Protocol):
    """プログラムを実行するためのエグゼキューター."""

    def build(self) -> bool:
        """プログラムをビルドする.

        Args:

        Returns:
            bool: ビルドが成功したか
        """

    def execute(self, test_case: AtcoderTestCase) -> AtCoderTestResult:
        """プログラムを実行し、テスト結果を得る.

        Args:
            test_case (AtcoderTestCase): テストケース

        Returns:
            TestResult: テスト結果
        """


def get_default_program_executor(
    build_command: List[str], run_command: List[str]
) -> ProgramExecutor:
    """ProgramExecutorRepoの標準実装を返す."""
    return ProgramExecutorRepoImpl(build_command, run_command)


class ProgramExecutorRepoImpl:
    """プログラム実行のためのロジックを集約させるためのリポジトリ実装."""

    _build_command: List[str]
    _run_command: List[str]

    def __init__(self, build_command: List[str], run_command: List[str]):
        """__init__."""
        self._build_command = build_command
        self._run_command = run_command

    def build(self) -> bool:
        """プログラムをビルドする.

        Args:

        Returns:
            bool: ビルドが成功したか
        """
        completed_process = subprocess.run(self._build_command)
        return completed_process.returncode == 0

    def execute(self, test_case: AtcoderTestCase) -> AtCoderTestResult:
        """プログラムを実行し、テスト結果を得る.

        Args:
            test_case (AtcoderTestCase): テストケース

        Returns:
            TestResult: テスト結果
        """
        completed_process = subprocess.run(
            self._run_command, input=test_case.given, text=True, capture_output=True
        )

        if completed_process.returncode != 0:
            return AtCoderTestResult(
                test_case.name,
                AtCoderTestStatus.ERROR,
                actual=completed_process.stdout,
                error=completed_process.stderr,
            )

        if test_case.expected is None:
            return AtCoderTestResult(
                test_case.name,
                AtCoderTestStatus.JUSTSHOW,
                actual=completed_process.stdout,
                error="",
            )

        if test_case.expected.rstrip() == completed_process.stdout.rstrip():
            return AtCoderTestResult(
                test_case.name,
                AtCoderTestStatus.AC,
                actual=completed_process.stdout,
                error="",
            )
        else:
            return AtCoderTestResult(
                test_case.name,
                AtCoderTestStatus.WA,
                actual=completed_process.stdout,
                expected=test_case.expected,
                error="",
            )



"""atcoderからlogin済みのセッションを取得するrepository."""


from typing import Protocol

import requests
from bs4 import BeautifulSoup

from atcoder_helper.repositories.errors import ConnectionError
from atcoder_helper.repositories.errors import LoginFailure
from atcoder_helper.repositories.errors import ParseError
from atcoder_helper.repositories.utils import AtCoderURLProvider


class AtCoderLoggedInSessionRepository(Protocol):
    """atcoder からlogin済みのセッションを取得するrepositoryのプロトコル."""

    def read(self, username: str, password: str) -> requests.Session:
        """atcoderにloginしたsessionを返す.入力したユーザーネームとパスワードは保存されず、代わりにセッションが保存される.

        Args:
            username (str): username
            password (str): password

        Raises:
            ConnectionError: GETかPOSTに失敗
            ParseError: パースに失敗
            LoginFailure: ログイン失敗

        Returns:
            requests.Session: セッションを返す.
        """


def get_default_atcoder_session_repository() -> AtCoderLoggedInSessionRepository:
    """AtCoderLoggedInSessionRepositoryのデフォルト実装を返す.

    Returns:
        AtCoderLoggedInSessionRepository: デフォルト実装
    """
    return AtCoderLoggedInSessionRepositoryImpl()


class AtCoderLoggedInSessionRepositoryImpl:
    """atcoder からlogin済みのセッションを取得するrepository."""

    _url_provider = AtCoderURLProvider

    def _get_csrf_token(self, session: requests.Session) -> str:
        """_.

        Args:
            session (requests.Session): _description_

        Raises:
            ConnectionError: _description_
            ParseError: _description_

        Returns:
            str: _description_
        """
        try:
            login_page = session.get(self._url_provider.login_url)
        except Exception as e:
            raise ConnectionError("fail to get url") from e

        try:
            html = BeautifulSoup(login_page.text, "html.parser")
            token = html.find("input").attrs["value"]
            if not isinstance(token, str):
                raise ParseError("input value is not a string")
        except Exception as e:
            raise ParseError("input value cannot be parsed") from e

        return token

    def read(self, username: str, password: str) -> requests.Session:
        """atcoderにloginしたsessionを返す.入力したユーザーネームとパスワードは保存されず、代わりにセッションが保存される.

        Args:
            username (str): username
            password (str): password

        Raises:
            ConnectionError: GETかPOSTに失敗
            ParseError: パースに失敗
            LoginFailure: ログイン失敗

        Returns:
            requests.Session: セッションを返す.
        """
        session = requests.Session()
        csrf_token = self._get_csrf_token(session)

        try:
            res = session.post(
                self._url_provider.login_url,
                params={
                    "username": username,
                    "password": password,
                    "csrf_token": csrf_token,
                },
                allow_redirects=False,
            )
        except Exception as e:
            raise ConnectionError(
                f"cannot post to {self._url_provider.login_url}"
            ) from e

        if res.headers["Location"] == "/home":
            return session
        else:
            raise LoginFailure("ログインに失敗しました.")



"""AtCoderからテストケースを取得するリポジトリ."""


from typing import List
from typing import Protocol

import requests
from bs4 import BeautifulSoup

from atcoder_helper.models.atcoder_test_case import AtcoderTestCase
from atcoder_helper.repositories.errors import ParseError
from atcoder_helper.repositories.utils import AtCoderURLProvider


class AtCoderTestCaseRepository(Protocol):
    """AtCoderからテストケースを取得するリポジトリのプロトコル."""

    def fetch_test_cases(
        self, session: requests.Session, contest: str, task: str
    ) -> List[AtcoderTestCase]:
        """テストケーススイートを取得する.

        Args:
            contest (str): コンテスト名
            task (str): タスク名

        Raises:
            ConnectionError: GETに失敗
            ParseError: Parseに失敗

        Returns:
            List[TestCase]: テストケーススイート
        """


def get_default_atcoder_test_case_repository() -> AtCoderTestCaseRepository:
    """AtCoderTestCaseRepositoryのデフォルト実装を作成する.

    Returns:
        AtCoderTestCaseRepository: デフォルト実装
    """
    return AtCoderTestCaseRepositoryImpl()


class AtCoderTestCaseRepositoryImpl:
    """AtCoderからテストケースを取得するリポジトリ."""

    _url_provider = AtCoderURLProvider

    def fetch_test_cases(
        self, session: requests.Session, contest: str, task: str
    ) -> List[AtcoderTestCase]:
        """テストケーススイートを取得する.

        Args:
            contest (str): コンテスト名
            task (str): タスク名

        Raises:
            ConnectionError: GETに失敗
            ParseError: Parseに失敗

        Returns:
            List[TestCase]: テストケーススイート
        """

        def normalize_newline(text: str) -> str:
            return "\n".join(text.splitlines())

        try:
            task_page = session.get(self._url_provider.task_url(contest, task))
        except Exception as e:
            raise ConnectionError(
                f"cannot GET {self._url_provider.task_url(contest, task)}"
            ) from e

        try:
            html = BeautifulSoup(task_page.text, "html.parser")

            sections = (
                html.find("div", id="task-statement")
                .find("span", attrs={"class": "lang-ja"})
                .find_all("section")
            )

            input_sections = {
                section.find("h3").text.split()[1]: normalize_newline(
                    section.find("pre").text
                )
                for section in sections
                if "入力例" in section.find("h3").text
            }

            output_sections = {
                section.find("h3").text.split()[1]: normalize_newline(
                    section.find("pre").text
                )
                for section in sections
                if "出力例" in section.find("h3").text
            }
        except Exception as e:
            raise ParseError() from e

        return [
            AtcoderTestCase(
                name=f"case-{name}", given=given, expected=output_sections[name]
            )
            for (name, given) in input_sections.items()
        ]



"""atcoder_helperコマンドのエントリポイント."""

from atcoder_helper.scripts.executor import get_default_executor
from atcoder_helper.scripts.parser import get_root_parser


def main() -> None:
    """entrypoint."""
    executor = get_default_executor()

    root_parser = get_root_parser()
    args = root_parser.parse_args()

    if hasattr(args, "handler"):
        args.handler(self=executor, args=args)
    else:
        args.parser.print_help()



"""SessionRepo."""

import os
import pickle
from typing import Final
from typing import Protocol
from typing import cast

import requests

from atcoder_helper.repositories.errors import ReadError
from atcoder_helper.repositories.errors import WriteError


class LoggedInSessionRepository(Protocol):
    """Login済みセッションのリポジトリのプロトコル."""

    def write(self, session: requests.Session) -> None:
        """_write.

        Raises:
            WriteError: 書き込みに失敗
        """

    def read(self) -> requests.Session:
        """read. ファイルが存在しない場合はデフォルトsessionを返す.

        Raises:
            ReadError: 読み込みに失敗した

        Return:
            requests.Session: 取得したsession
        """

    def doesExist(self) -> bool:
        """sessionがすでに存在するかどうか確認する.

        Returns:
            bool: 存在するかどうか
        """

    def delete(self) -> None:
        """Session 情報を削除します.

        Raises:
            WriteError: _
        """


def get_default_session_repository() -> LoggedInSessionRepository:
    """LoggedInSessionRepositoryの標準実装を返す.

    Returns:
        LoggedInSessionRepositoryImpl: 標準実装
    """
    default_session_file: Final[str] = os.path.join(
        os.path.expanduser("~"), ".atcoder_helper", "session", "session_dump.pkl"
    )
    return LoggedInSessionRepositoryImpl(default_session_file)


class LoggedInSessionRepositoryImpl:
    """Login済みセッションの永続化層."""

    _session_filename: str

    def __init__(self, session_filename: str):
        """__init__.

        Args:
            session_filename (str): session filename
        """
        self._session_filename = session_filename

    def write(self, session: requests.Session) -> None:
        """_write.

        Raises:
            WriteError: 書き込みに失敗
        """
        os.makedirs(os.path.dirname(self._session_filename), exist_ok=True)

        try:
            with open(self._session_filename, "wb") as file:
                pickle.dump(session, file)
        except OSError as e:
            raise WriteError(f"cannot write to {self._session_filename}") from e

    def read(self) -> requests.Session:
        """read. ファイルが存在しない場合はデフォルトsessionを返す.

        Raises:
            ReadError: 読み込みに失敗した

        Return:
            requests.Session: 取得したsession
        """
        if self.doesExist():
            try:
                with open(self._session_filename, "rb") as file:
                    return cast(requests.Session, pickle.load(file))
            except OSError as e:
                raise ReadError(f"cannot open or parse {self._session_filename}") from e
        else:
            return requests.session()

    def doesExist(self) -> bool:
        """sessionがすでに存在するかどうか確認する.

        Returns:
            bool: 存在するかどうか
        """
        return os.path.isfile(self._session_filename)

    def delete(self) -> None:
        """Session 情報を削除します.

        Raises:
            WriteError: _
        """
        self.write(requests.Session())



"""atcoder_helperコマンドの実行時情報を保持するクラス."""
import argparse
import getpass
import os
import sys
import traceback

from atcoder_helper.services import errors as service_errors
from atcoder_helper.services.atcoder_helper_config import AtCoderHelperConfigService
from atcoder_helper.services.atcoder_helper_config import (
    get_default_atcoder_helper_config_service,
)
from atcoder_helper.services.auth import AuthService
from atcoder_helper.services.auth import get_default_auth_service
from atcoder_helper.services.execute_test import ExecuteTestService
from atcoder_helper.services.execute_test import get_default_execute_test_service
from atcoder_helper.services.fetch_task import FetchTaskService
from atcoder_helper.services.fetch_task import get_default_fetch_task_service
from atcoder_helper.services.init_task import InitTaskDirService
from atcoder_helper.services.init_task import get_default_init_task_dir_service


class Executor:
    """実行時に必要な情報を持ちまわるためのクラス."""

    _auth_service: AuthService
    _atcoder_helper_config_service: AtCoderHelperConfigService
    _execute_test_service: ExecuteTestService
    _fetch_task_service: FetchTaskService
    _init_task_dir_service: InitTaskDirService

    def __init__(
        self,
        auth_service: AuthService,
        atcoder_helper_config_service: AtCoderHelperConfigService,
        execute_test_service: ExecuteTestService,
        fetch_task_service: FetchTaskService,
        init_task_dir_service: InitTaskDirService,
    ) -> None:
        """__init__.

        Args:
            auth_service (AuthService, optional): _
            atcoder_helper_config_service (AtCoderHelperConfigService, optional): _
            execute_test_service (ExecuteTestService, optional): _
            fetch_task_service (FetchTaskService, optional): _
            init_task_dir_service (InitTaskDirService, optional): _
        """
        self._auth_service = auth_service
        self._atcoder_helper_config_service = atcoder_helper_config_service
        self._execute_test_service = execute_test_service
        self._fetch_task_service = fetch_task_service
        self._init_task_dir_service = init_task_dir_service

    def auth_login_handler(self, args: argparse.Namespace) -> None:
        """ログインする."""
        name = input("name:")
        password = getpass.getpass("password:")
        try:
            self._auth_service.login(name, password)
        except service_errors.AlreadyLoggedIn:
            print("既にログインしています.")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)
        except service_errors.ConfigAccessError:
            print("設定ファイルを正しく読み込めません.")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)
        except service_errors.AtcoderAccessError:
            print("AtCoderのサイトと正しく通信ができません.")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)

        print("logged in.")

    def auth_logout_handler(self, args: argparse.Namespace) -> None:
        """ログアウトする.

        Args:
            args (argparse.Namespace): _description_
        """
        try:
            self._auth_service.logout()
        except service_errors.ConfigAccessError:
            print("設定ファイルを正しく読み込めません.")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)

    def auth_status_handler(self, args: argparse.Namespace) -> None:
        """現在ログインしているかどうかを確認する."""
        try:
            stat = self._auth_service.status()
        except service_errors.AtcoderAccessError:
            print("atcoderと通信できません")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)
        except service_errors.ConfigAccessError:
            print("設定ファイルが読み込めません")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)

        if stat:
            print("logged in.")
        else:
            print("logged out.")

    def task_init_handler(self, args: argparse.Namespace) -> None:
        """既存のディレクトリをタスク用に初期化する.

        Args:
            args (argparse.Namespace): 引数
        """
        try:
            self._init_task_dir_service.init_task()
        except service_errors.ConfigAccessError:
            print("設定ファイルへの読み書きでエラーが発生しました")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)
        except service_errors.DirectoryNotEmpty:
            print("初期化しようとしているディレクトリが空ではありません")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)

    def task_create_handler(self, args: argparse.Namespace) -> None:
        """新規にディレクトリを作り、タスク用に初期化する.

        Args:
            args (argparse.Namespace): コマンドライン引数
        """
        try:
            self._init_task_dir_service.init_task(
                dir=os.path.join(args.contest, args.task),
                contest=args.contest,
                task=args.task,
            )
        except service_errors.ConfigAccessError:
            print("設定ファイルへの読み書きでエラーが発生しました")
            if args.verbose:
                print(traceback.format_exc())
        except service_errors.DirectoryNotEmpty:
            print("初期化しようとしているディレクトリが空ではありません")
            if args.verbose:
                print(traceback.format_exc())

    def execute_test_handler(self, args: argparse.Namespace) -> None:
        """テストスイートを実行する.

        Args:
            args (argparse.Namespace): コマンドライン引数
        """
        try:
            self._execute_test_service.execute_test()
        except service_errors.ConfigAccessError:
            print("設定ファイルの読み書きに失敗しました")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)

    def fetch_task_handler(self, args: argparse.Namespace) -> None:
        """テストケースをフェッチする.

        Args:
            args (argparse.Namespace): コマンドライン引数
        """
        try:
            self._fetch_task_service.fetch_task(args.contest, args.task)
        except service_errors.AtcoderAccessError:
            print("AtCoderサイトからデータを取得する過程でエラーが発生しました")
            if args.verbose:
                print(traceback.format_exc())

            sys.exit(1)

    def config_init_handler(self, args: argparse.Namespace) -> None:
        """atcoder_helper全体設定ファイルを初期化する.

        Args:
            args (argparse.Namespace): 引数
        """
        try:
            self._atcoder_helper_config_service.init_config()
        except service_errors.ConfigAccessError:
            print("設定ファイルの読み書きに失敗しました")
            if args.verbose:
                print(traceback.format_exc())

            sys.exit(1)

    def config_languages_handler(self, args: argparse.Namespace) -> None:
        """使用可能な言語一覧を表示する.

        Args:
            args (argparse.Namespace): 引数
        """
        try:
            languages = self._atcoder_helper_config_service.config_languages()
        except service_errors.ConfigAccessError:
            print("設定ファイルの読み書きに失敗しました")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)

        for language_name in languages:
            print(language_name)

    def config_default_language_handler(self, args: argparse.Namespace) -> None:
        """デフォルトの言語を表示する.

        Args:
            args (argparse.Namespace): 引数
        """
        try:
            default_language = (
                self._atcoder_helper_config_service.config_default_language()
            )
        except service_errors.ConfigAccessError:
            print("設定ファイルの読み書きに失敗しました")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)

        print(default_language.name)

    def config_use_handler(self, args: argparse.Namespace) -> None:
        """デフォルトの言語を設定する.

        Args:
            args (argparse.Namespace): 引数
        """
        try:
            self._atcoder_helper_config_service.config_use(args.language)
        except service_errors.UndefinedLanguage:
            print(f"{args.language}は使用可能な言語の中に存在しません。設定ファイルを変更し、言語設定を追加してください。")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)
        except service_errors.ConfigAccessError:
            print("設定ファイルの読み書きに失敗しました")
            if args.verbose:
                print(traceback.format_exc())
            sys.exit(1)


def get_default_executor() -> Executor:
    """標準Executorを返す."""
    return Executor(
        auth_service=get_default_auth_service(),
        atcoder_helper_config_service=(get_default_atcoder_helper_config_service()),
        execute_test_service=get_default_execute_test_service(),
        fetch_task_service=get_default_fetch_task_service(),
        init_task_dir_service=get_default_init_task_dir_service(),
    )



"""Tests for Executor."""


import argparse
from typing import Any
from typing import Optional

import mock
import pytest

from atcoder_helper.models.atcoder_helper_config import LanguageConfig
from atcoder_helper.scripts.executor import Executor
from atcoder_helper.services.atcoder_helper_config import AtCoderHelperConfigService
from atcoder_helper.services.auth import AuthService
from atcoder_helper.services.errors import AlreadyLoggedIn
from atcoder_helper.services.errors import AtcoderAccessError
from atcoder_helper.services.errors import ConfigAccessError
from atcoder_helper.services.errors import DirectoryNotEmpty
from atcoder_helper.services.errors import UndefinedLanguage
from atcoder_helper.services.execute_test import ExecuteTestService
from atcoder_helper.services.fetch_task import FetchTaskService
from atcoder_helper.services.init_task import InitTaskDirService


def _get_sut(
    auth_service_mock: AuthService = mock.MagicMock(),
    atcoder_helper_config_service_mock: AtCoderHelperConfigService = mock.MagicMock(),
    execute_test_service_mock: ExecuteTestService = mock.MagicMock(),
    fetch_task_service_mock: FetchTaskService = mock.MagicMock(),
    init_task_dir_service_mock: InitTaskDirService = mock.MagicMock(),
) -> Executor:
    return Executor(
        auth_service=auth_service_mock,
        atcoder_helper_config_service=atcoder_helper_config_service_mock,
        execute_test_service=execute_test_service_mock,
        fetch_task_service=fetch_task_service_mock,
        init_task_dir_service=init_task_dir_service_mock,
    )


def _get_default_namespace(**additional_kwargs: str) -> argparse.Namespace:
    return argparse.Namespace(verbose=False, **additional_kwargs)


@pytest.mark.parametrize(
    argnames=(
        "login_side_effect",
        "should_succeed",
    ),
    argvalues=[
        [None, True],
        [AlreadyLoggedIn(), False],
        [ConfigAccessError(), False],
        [AtcoderAccessError(), False],
    ],
)
def test_auth_login_handler(
    login_side_effect: Optional[Exception], should_succeed: bool
) -> None:
    """auth_login_handlerのテスト."""
    sut = _get_sut(
        auth_service_mock=mock.MagicMock(
            login=mock.MagicMock(side_effect=login_side_effect)
        )
    )

    name = "foo"
    password = "bar"

    with (
        mock.patch("builtins.input", return_value=name),
        mock.patch("getpass.getpass", return_value=password),
    ):
        if should_succeed:
            sut.auth_login_handler(_get_default_namespace())
        else:
            with pytest.raises(SystemExit) as e:
                sut.auth_login_handler(_get_default_namespace())
            assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("logout_side_effect", "should_succeed"),
    argvalues=[[None, True], [ConfigAccessError(), False]],
)
def test_auth_logout_handler(
    logout_side_effect: Exception, should_succeed: bool
) -> None:
    """auth_logout_handlerのテスト."""
    sut = _get_sut(
        auth_service_mock=mock.MagicMock(
            logout=mock.MagicMock(side_effect=logout_side_effect)
        )
    )
    if should_succeed:
        sut.auth_logout_handler(_get_default_namespace())
    else:
        with pytest.raises(SystemExit) as e:
            sut.auth_logout_handler(_get_default_namespace())
        assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("status_return_value", "status_side_effect", "message", "should_succeed"),
    argvalues=[
        [True, None, "logged in.\n", True],
        [False, None, "logged out.\n", True],
        [False, AtcoderAccessError(), None, False],
    ],
)
def test_auth_status_handler(
    capfd: Any,
    status_return_value: bool,
    status_side_effect: Exception,
    message: str,
    should_succeed: bool,
) -> None:
    """auth_status_handlerのテスト."""
    sut = _get_sut(
        auth_service_mock=mock.MagicMock(
            status=mock.MagicMock(
                side_effect=status_side_effect, return_value=status_return_value
            )
        )
    )

    if should_succeed:
        sut.auth_status_handler(_get_default_namespace())
        assert capfd.readouterr().out == message
    else:
        with pytest.raises(SystemExit) as e:
            sut.auth_status_handler(_get_default_namespace())
        assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("init_task_side_effect", "should_succeed"),
    argvalues=[[None, True], [ConfigAccessError, False], [DirectoryNotEmpty, False]],
)
def test_task_init_handler(
    init_task_side_effect: Exception, should_succeed: bool
) -> None:
    """task_init_handlerのテスト."""
    sut = _get_sut(
        init_task_dir_service_mock=mock.MagicMock(
            init_task=mock.MagicMock(side_effect=init_task_side_effect)
        )
    )

    if should_succeed:
        sut.task_init_handler(_get_default_namespace())
    else:
        with pytest.raises(SystemExit) as e:
            sut.task_init_handler(_get_default_namespace())
        assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("init_task_side_effect", "should_succeed"),
    argvalues=[
        [None, True],
        [ConfigAccessError(), False],
        [DirectoryNotEmpty(), False],
    ],
)
def test_task_create_handler(
    init_task_side_effect: Exception, should_succeed: bool
) -> None:
    """task_create_handlerのテスト."""
    contest = "foo"
    task = "bar"
    sut = _get_sut(
        init_task_dir_service_mock=mock.MagicMock(
            init_task=mock.MagicMock(side_effect=init_task_side_effect)
        )
    )

    namespace = _get_default_namespace(contest=contest, task=task)

    if should_succeed:
        sut.task_init_handler(namespace)
    else:
        with pytest.raises(SystemExit) as e:
            sut.task_init_handler(namespace)
        assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("execute_test_side_effect", "should_succeed"),
    argvalues=[[None, True], [ConfigAccessError(), False]],
)
def test_execute_test_handler(
    execute_test_side_effect: Exception, should_succeed: bool
) -> None:
    """execute_test_handlerのテスト."""
    sut = _get_sut(
        execute_test_service_mock=mock.MagicMock(
            execute_test=mock.MagicMock(side_effect=execute_test_side_effect)
        )
    )

    if should_succeed:
        sut.execute_test_handler(_get_default_namespace())
    else:
        with pytest.raises(SystemExit) as e:
            sut.execute_test_handler(_get_default_namespace())
        assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("fetch_task_side_effect", "should_succeed"),
    argvalues=[[None, True], [AtcoderAccessError, False]],
)
def test_fetch_task_handler(
    fetch_task_side_effect: Exception, should_succeed: bool
) -> None:
    """fetch_task_handlerのテスト."""
    sut = _get_sut(
        fetch_task_service_mock=mock.MagicMock(
            fetch_task=mock.MagicMock(side_effect=fetch_task_side_effect)
        )
    )
    contest = "foo"
    task = "bar"
    args = _get_default_namespace(contest=contest, task=task)

    if should_succeed:
        sut.fetch_task_handler(args)
    else:
        with pytest.raises(SystemExit) as e:
            sut.fetch_task_handler(args)
        assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("init_config_side_effect", "should_succeed"),
    argvalues=[[None, True], [ConfigAccessError, False]],
)
def test_config_init_handler(
    init_config_side_effect: Exception, should_succeed: bool
) -> None:
    """config_init_handlerのテスト."""
    sut = _get_sut(
        atcoder_helper_config_service_mock=mock.MagicMock(
            init_config=mock.MagicMock(side_effect=init_config_side_effect)
        )
    )

    args = _get_default_namespace()

    if should_succeed:
        sut.config_init_handler(args)
    else:
        with pytest.raises(SystemExit) as e:
            sut.config_init_handler(args)
        assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("config_languages_side_effect", "should_succeed"),
    argvalues=[[None, True], [ConfigAccessError, False]],
)
def test_config_languages_handler(
    config_languages_side_effect: Exception, should_succeed: bool
) -> None:
    """config_languages_handlerのテスト."""
    languages = {"c": LanguageConfig(name="c", build=[], run=[])}

    sut = _get_sut(
        atcoder_helper_config_service_mock=mock.MagicMock(
            config_languages=mock.MagicMock(
                side_effect=config_languages_side_effect, return_value=languages
            )
        )
    )

    args = _get_default_namespace()

    if should_succeed:
        sut.config_languages_handler(args)
    else:
        with pytest.raises(SystemExit) as e:
            sut.config_languages_handler(args)
        assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("default_language_side_effect", "should_succeed"),
    argvalues=[[None, True], [ConfigAccessError(), False]],
)
def test_config_default_language_handler(
    default_language_side_effect: Exception, should_succeed: bool
) -> None:
    """config_default_language_handlerのテスト."""
    language = LanguageConfig(name="c", build=[], run=[])

    sut = _get_sut(
        atcoder_helper_config_service_mock=mock.MagicMock(
            config_default_language=mock.MagicMock(
                side_effect=default_language_side_effect, return_value=language
            )
        )
    )

    args = _get_default_namespace()

    if should_succeed:
        sut.config_default_language_handler(args)
    else:
        with pytest.raises(SystemExit) as e:
            sut.config_default_language_handler(args)
        assert e.value.code == 1


@pytest.mark.parametrize(
    argnames=("config_use_side_effect", "should_succeed"),
    argvalues=[
        [None, True],
        [UndefinedLanguage(), False],
        [ConfigAccessError(), False],
    ],
)
def test_config_use_handler(
    config_use_side_effect: Exception, should_succeed: bool
) -> None:
    """config_use_handlerのテスト."""
    language = "csharp"

    sut = _get_sut(
        atcoder_helper_config_service_mock=mock.MagicMock(
            config_use=mock.MagicMock(side_effect=config_use_side_effect)
        )
    )

    args = _get_default_namespace(language=language)

    if should_succeed:
        sut.config_use_handler(args)
    else:
        with pytest.raises(SystemExit) as e:
            sut.config_use_handler(args)
        assert e.value.code == 1



"""testcasesを取得し、指定したtestcasesファイルに書き込むサービス."""
from typing import Optional
from typing import Protocol
from typing import Tuple

from atcoder_helper.repositories import errors as repository_error
from atcoder_helper.repositories.atcoder_test_case_repo import AtCoderTestCaseRepository
from atcoder_helper.repositories.atcoder_test_case_repo import (
    get_default_atcoder_test_case_repository,
)
from atcoder_helper.repositories.local_test_case_repo import LocalTestCaseRepository
from atcoder_helper.repositories.local_test_case_repo import (
    get_default_local_test_case_repository,
)
from atcoder_helper.repositories.logged_in_session_repo import LoggedInSessionRepository
from atcoder_helper.repositories.logged_in_session_repo import (
    get_default_session_repository,
)
from atcoder_helper.repositories.task_config_repo import TaskConfigRepository
from atcoder_helper.repositories.task_config_repo import (
    get_default_task_config_repository,
)
from atcoder_helper.services.errors import AtcoderAccessError
from atcoder_helper.services.errors import ConfigAccessError


class FetchTaskService(Protocol):
    """atcoderサイトからテストケースを取得するサービスのプロトコル."""

    def fetch_task(
        self,
        contest: Optional[str],
        task: Optional[str],
    ) -> None:
        """testcasesを取得し、指定したtestcasesファイルに書き込む.

        Raises:
            AtcoderAccessError: atcoderとのデータのやりとりの中でのエラー
            ConfigAccessError: 設定ファイルの読み書きのエラー

        Args:
            contest (str): コンテスト名。AtCoderのコンテストページのURLに現れる形式で渡す
            task (str): タスク名。AtCoderのコンテストページのURLに現れる形から、"コンテスト名_"の部分を除いたもの
        """


def get_default_fetch_task_service() -> FetchTaskService:
    """FetchTaskServiceの標準実装を返す.

    Returns:
        FetchTaskService:
    """
    return FetchTaskServiceImpl(
        task_config_repo=get_default_task_config_repository(),
        test_case_repo=get_default_local_test_case_repository(),
        session_repo=get_default_session_repository(),
        atcoder_testcase_repo=(get_default_atcoder_test_case_repository()),
    )


class FetchTaskServiceImpl:
    """atcoderサイトからテストケースを取得するサービス."""

    _task_config_repo: TaskConfigRepository
    _test_case_repo: LocalTestCaseRepository
    _session_repo: LoggedInSessionRepository
    _atcoder_testcase_repo: AtCoderTestCaseRepository

    def __init__(
        self,
        task_config_repo: TaskConfigRepository,
        test_case_repo: LocalTestCaseRepository,
        session_repo: LoggedInSessionRepository,
        atcoder_testcase_repo: AtCoderTestCaseRepository,
    ) -> None:
        """__init__."""
        self._task_config_repo = task_config_repo
        self._test_case_repo = test_case_repo
        self._session_repo = session_repo
        self._atcoder_testcase_repo = atcoder_testcase_repo

    def fetch_task(
        self,
        contest: Optional[str],
        task: Optional[str],
    ) -> None:
        """testcasesを取得し、指定したtestcasesファイルに書き込む.

        Raises:
            AtcoderAccessError: atcoderとのデータのやりとりの中でのエラー
            ConfigAccessError: 設定ファイルの読み書きのエラー

        Args:
            contest (str): コンテスト名。AtCoderのコンテストページのURLに現れる形式で渡す
            task (str): タスク名。AtCoderのコンテストページのURLに現れる形から、"コンテスト名_"の部分を除いたもの
        """
        (contest, task) = self._get_task_info(contest, task)

        try:
            session = self._session_repo.read()
        except repository_error.ReadError as e:
            raise ConfigAccessError("セッションの読み込みに失敗") from e

        try:
            test_cases = self._atcoder_testcase_repo.fetch_test_cases(
                session=session, contest=contest, task=task
            )
        except (repository_error.ConnectionError, repository_error.ParseError) as e:
            raise AtcoderAccessError("テストケースの取得に失敗しました") from e

        try:
            self._test_case_repo.write(test_cases)
        except repository_error.WriteError as e:
            raise ConfigAccessError("テストケースの書き込みに失敗しました") from e

    def _get_task_info(
        self, given_contest: Optional[str], given_task: Optional[str]
    ) -> Tuple[str, str]:
        """contestとtaskを得る.

        Raises:
            ConfigAccessError: 設定に問題がある

        Returns:
            Tuple[str, str]: contest, task
        """
        try:
            task_config = self._task_config_repo.read()
        except (repository_error.ReadError, repository_error.ParseError) as e:
            raise ConfigAccessError("設定ファイルの読み込みに失敗しました") from e

        contest = task_config.contest if given_contest is None else given_contest
        task = task_config.task if given_task is None else given_task

        if contest is None:
            raise ConfigAccessError("contest is not set.")

        if task is None:
            raise ConfigAccessError("task is not set.")

        return (contest, task)



"""テストケース実行のためのメソッド."""

from textwrap import indent
from typing import Any
from typing import Callable
from typing import List
from typing import Protocol

from atcoder_helper.models.atcoder_test_case import AtCoderTestResult
from atcoder_helper.models.atcoder_test_case import AtCoderTestStatus
from atcoder_helper.program_executor import ProgramExecutor
from atcoder_helper.program_executor import get_default_program_executor
from atcoder_helper.repositories import errors as repository_error
from atcoder_helper.repositories.local_test_case_repo import LocalTestCaseRepository
from atcoder_helper.repositories.local_test_case_repo import (
    get_default_local_test_case_repository,
)
from atcoder_helper.repositories.task_config_repo import TaskConfigRepository
from atcoder_helper.repositories.task_config_repo import (
    get_default_task_config_repository,
)
from atcoder_helper.services.errors import ConfigAccessError

ExecutorBuilder = Callable[[List[str], List[str]], ProgramExecutor]


class ExecuteTestService(Protocol):
    """テストを実行するサービス."""

    def execute_test(self) -> None:
        """testcaseに基づき、テストを実行する関数.

        Raises:
            ConfigAccessError: 設定ファイル読み書きのエラー
        """


def get_default_execute_test_service() -> ExecuteTestService:
    """ExecuteTestServiceの標準実装を返す.

    Returns:
        ExecuteTestService:
    """
    return ExecuteTestServiceImpl(
        task_config_repo=get_default_task_config_repository(),
        test_case_repo=get_default_local_test_case_repository(),
        executor_builder=get_default_program_executor,
    )


class ExecuteTestServiceImpl:
    """テストを実行するサービス."""

    _task_config_repo: TaskConfigRepository
    _test_case_repo: LocalTestCaseRepository

    # 本当は ExecutorBuilder型なんだがmypyのバグにより型付けに失敗するので Any
    # see also https://github.com/python/mypy/issues/5485
    _executor_builder: Any

    def __init__(
        self,
        task_config_repo: TaskConfigRepository,
        test_case_repo: LocalTestCaseRepository,
        executor_builder: ExecutorBuilder,
    ):
        """__init__.

        Args:
            task_config_repo (TaskConfigRepository, optional): _
            test_case_repo (TestCaseRepository, optional): _
            executor_builder (Callable[[List[str], List[str]], ProgramExecutor]): _
        """
        self._task_config_repo = task_config_repo
        self._test_case_repo = test_case_repo
        self._executor_builder = executor_builder

    def execute_test(self) -> None:
        """testcaseに基づき、テストを実行する関数.

        Raises:
            ConfigAccessError: 設定ファイル読み書きのエラー
        """
        try:
            task_config = self._task_config_repo.read()
            test_cases = self._test_case_repo.read()
        except (repository_error.ReadError, repository_error.ParseError):
            raise ConfigAccessError("設定ファイルの読み込みに失敗しました")

        executor = self._executor_builder(task_config.build, task_config.run)

        executor.build()  # TODO(ビルド失敗で止まるようにする)

        results = []
        for test_case in test_cases:
            print("-----------------------------------")
            print(f"executing {test_case.name}...")
            result = executor.execute(test_case)
            results.append(result)
            self._show_result(result)
        self._show_summary(results)

    def _show_result(self, result: AtCoderTestResult) -> None:
        print("-----------------------------------")
        print(f"{result.name:<15}: {result.status.dyed}")

        if result.status == AtCoderTestStatus.JUSTSHOW:
            print("    output:")
            print(indent(result.actual, "       >"))
        if result.status == AtCoderTestStatus.ERROR:
            print(result.error)
        if result.status == AtCoderTestStatus.WA:
            if result.expected is None:
                raise Exception("internal error")

            print("    expected:")
            print(indent(result.expected, "       >"))
            print("    but got:")
            print(indent(result.actual, "       >"))

    def _show_summary(self, results: List[AtCoderTestResult]) -> None:
        print("========================================")
        print("SUMMARY:")
        for result in results:
            print(f"{result.name:<15}: {result.status.dyed}")



"""Repository Utilitiles."""


from typing import Final


class AtCoderURLProvider:
    """URL情報を保持するクラス."""

    _atcoder_url: Final[str] = "https://atcoder.jp"
    login_url: Final[str] = f"{_atcoder_url}/login"

    @staticmethod
    def contest_url(contest: str) -> str:
        """contestのURLを返す.

        Args:
            contest (str):

        Returns:
            str:
        """
        return f"{AtCoderURLProvider._atcoder_url}/contests/{contest}"

    @staticmethod
    def task_url(contest: str, task: str) -> str:
        """taskのURLを返す.

        Args:
            contest (str):
            task (str):

        Returns:
            str:
        """
        return f"{AtCoderURLProvider.contest_url(contest)}/tasks/{contest}_{task}"

    @staticmethod
    def submit_url(contest: str) -> str:
        """submitのURLを返す.

        Args:
            contest (str):

        Returns:
            str:
        """
        return f"{AtCoderURLProvider.contest_url(contest)}/submit"



"""認証周りのサービス."""


from typing import Protocol

import atcoder_helper.repositories.errors as repository_error
from atcoder_helper.repositories.atcoder_logged_in_session_repo import (
    AtCoderLoggedInSessionRepository,
)
from atcoder_helper.repositories.atcoder_logged_in_session_repo import (
    get_default_atcoder_session_repository,
)
from atcoder_helper.repositories.logged_in_session_repo import LoggedInSessionRepository
from atcoder_helper.repositories.logged_in_session_repo import (
    get_default_session_repository,
)
from atcoder_helper.repositories.login_status_repo import LoginStatusRepo
from atcoder_helper.repositories.login_status_repo import get_default_login_status_repo
from atcoder_helper.services.errors import AtcoderAccessError
from atcoder_helper.services.errors import ConfigAccessError


class AuthService(Protocol):
    """auth を扱うサービスのプロトコル."""

    def login(self, username: str, password: str) -> None:
        """ログインする.

        Args:
            username (str): username
            password (str): password

        Raises:
            ConfigAccessError: 設定ファイルのエラー
            AtcoderAccessError: atcoderから情報を取得する際のエラー
            AlreadyLoggedIn: 既にログインしている # 今は返さないけど今後かなりの高確率で返しうるので書いておく
        """

    def logout(self) -> None:
        """logout.

        Raises:
            ConfigAccessError: 設定ファイル読み書きのエラー
        """

    def status(self) -> bool:
        """loginしているかどうかを返す.

        Returns:
            bool: loginしているか

        Raises:
            AtcoderAccessError: atcoder access error
        """


def get_default_auth_service() -> AuthService:
    """AuthServiceの標準実装を返す."""
    return AuthServiceImpl(
        atcoder_session_repo=(get_default_atcoder_session_repository()),
        local_session_repo=(get_default_session_repository()),
        login_status_repo=get_default_login_status_repo(),
    )


class AuthServiceImpl:
    """auth を扱うサービス."""

    _atcoder_session_repo: AtCoderLoggedInSessionRepository
    _local_session_repo: LoggedInSessionRepository
    _login_status_repo: LoginStatusRepo

    def __init__(
        self,
        atcoder_session_repo: AtCoderLoggedInSessionRepository,
        local_session_repo: LoggedInSessionRepository,
        login_status_repo: LoginStatusRepo,
    ):
        """__init__.

        Args:
            atcoder_repo (AtCoderRepository, optional): _
            atcoder_session_repo (AtCoderLoggedInSessionRepository, optional): _
            local_session_repo (LoggedInSessionRepository, optional): _
            login_status_repo (LoginStatusRepo, optional): _
        """
        self._atcoder_session_repo = atcoder_session_repo
        self._local_session_repo = local_session_repo
        self._login_status_repo = login_status_repo

    def login(self, username: str, password: str) -> None:
        """ログインする.

        Args:
            username (str): username
            password (str): password

        Raises:
            ConfigAccessError: 設定ファイルのエラー
            AtcoderAccessError: atcoderから情報を取得する際のエラー
            AlreadyLoggedInError: 既にログインしている
        """
        try:
            session = self._atcoder_session_repo.read(username, password)
        except repository_error.ConnectionError as e:
            raise AtcoderAccessError("通信に失敗しました") from e
        except repository_error.ParseError as e:
            raise AtcoderAccessError("パースに失敗しました.") from e
        except repository_error.LoginFailure as e:
            raise AtcoderAccessError("ログインに失敗しました") from e

        try:
            self._local_session_repo.write(session)
        except (repository_error.WriteError) as e:
            raise ConfigAccessError("セッションの保存に失敗しました") from e

    def logout(self) -> None:
        """logout.

        Raises:
            ConfigAccessError: 設定ファイル読み書きのエラー
        """
        try:
            self._local_session_repo.delete()
        except (repository_error.WriteError) as e:
            raise ConfigAccessError("設定ファイルの書き込みに失敗しました") from e

    def status(self) -> bool:
        """loginしているかどうかを返す.

        Returns:
            bool: loginしているか

        Raises:
            AtcoderAccessError: atcoder access error
            ConfigAccessError: config access error
        """
        try:
            session = self._local_session_repo.read()
        except repository_error.ReadError as e:
            raise ConfigAccessError("設定ファイルの読み込みに失敗しました.") from e

        try:
            status = self._login_status_repo.is_logged_in(session)
        except repository_error.ReadError as e:
            raise AtcoderAccessError("atcoderのページとの通信に失敗しました") from e

        return status



"""parser."""

import argparse

from atcoder_helper.scripts.executor import Executor


def get_root_parser() -> argparse.ArgumentParser:
    """rootParserを返します.

    Returns:
        argparse.ArgumentParser:
    """
    root_parser = argparse.ArgumentParser(description="atcoder の手助けをするコマンド")
    root_parser.set_defaults(parser=root_parser)
    root_parser.add_argument("--verbose", action="store_true")
    root_subparsers = root_parser.add_subparsers()

    _set_auth_parser(root_subparsers.add_parser("auth"))
    _set_exec_parser(root_subparsers.add_parser("exec"))
    _set_fetch_parser(root_subparsers.add_parser("fetch"))
    _set_task_pasrer(root_subparsers.add_parser("task"))
    _set_config_parser(root_subparsers.add_parser("config"))

    return root_parser


def _set_auth_parser(parser_auth: argparse.ArgumentParser) -> None:
    parser_auth.set_defaults(parser=parser_auth)

    parser_auth_subparsers = parser_auth.add_subparsers()

    parser_auth_login = parser_auth_subparsers.add_parser("login")
    parser_auth_login.set_defaults(
        handler=Executor.auth_login_handler, parser=parser_auth_login
    )

    parser_auth_logout = parser_auth_subparsers.add_parser("logout")
    parser_auth_logout.set_defaults(
        handler=Executor.auth_logout_handler, parser=parser_auth_logout
    )

    parser_auth_status = parser_auth_subparsers.add_parser("status")
    parser_auth_status.set_defaults(
        handler=Executor.auth_status_handler, parser=parser_auth_status
    )


def _set_exec_parser(parser_exec: argparse.ArgumentParser) -> None:
    parser_exec.set_defaults(handler=Executor.execute_test_handler, parser=parser_exec)


def _set_fetch_parser(parser_fetch: argparse.ArgumentParser) -> None:
    parser_fetch.set_defaults(handler=Executor.fetch_task_handler, parser=parser_fetch)
    parser_fetch.add_argument("--contest")
    parser_fetch.add_argument("--task")


def _set_task_pasrer(parser_task: argparse.ArgumentParser) -> None:
    parser_task.set_defaults(parser=parser_task)

    parser_task_subparsers = parser_task.add_subparsers()

    parser_task_init = parser_task_subparsers.add_parser("init")
    parser_task_init.set_defaults(
        handler=Executor.task_init_handler, parser=parser_task_init
    )

    parser_task_create = parser_task_subparsers.add_parser("create")
    parser_task_create.set_defaults(
        handler=Executor.task_create_handler, parser=parser_task_create
    )
    parser_task_create.add_argument("contest")
    parser_task_create.add_argument("task")


def _set_config_parser(parser_config: argparse.ArgumentParser) -> None:
    parser_config.set_defaults(parser=parser_config)

    parser_config_subparsers = parser_config.add_subparsers()

    parser_config_init = parser_config_subparsers.add_parser("init")
    parser_config_init.set_defaults(
        handler=Executor.config_init_handler, parser=parser_config_init
    )

    parser_config_languages = parser_config_subparsers.add_parser("languages")
    parser_config_languages.set_defaults(
        handler=Executor.config_languages_handler, parser=parser_config_languages
    )

    parser_config_default_language = parser_config_subparsers.add_parser("default")
    parser_config_default_language.set_defaults(
        handler=Executor.config_default_language_handler,
        parser=parser_config_default_language,
    )

    parser_config_use = parser_config_subparsers.add_parser("use")
    parser_config_use.set_defaults(
        handler=Executor.config_use_handler, parser=parser_config_use
    )
    parser_config_use.add_argument("language")



"""TaskConfigを定義する."""

from typing import List
from typing import Optional

from pydantic import BaseModel


class TaskConfig(BaseModel):
    """タスクごとの設定を保持する."""

    build: List[str]
    run: List[str]
    contest: Optional[str] = None
    task: Optional[str] = None



"""local_test_case_repoのテスト."""

from typing import Type

import mock
import pytest
import yaml
from mock import ANY
from pytest import MonkeyPatch

from atcoder_helper.models.atcoder_test_case import AtcoderTestCase
from atcoder_helper.repositories.errors import ParseError
from atcoder_helper.repositories.errors import ReadError
from atcoder_helper.repositories.errors import WriteError
from atcoder_helper.repositories.local_test_case_repo import LocalTestCaseRepositoryImpl

testcases = [AtcoderTestCase(name="foo_case", given="given")]


test_write_parameters = {
    "OK": [mock.MagicMock(), None],
    "Error(書き込み失敗)": [mock.MagicMock(side_effect=OSError()), WriteError],
}


@pytest.mark.parametrize(
    argnames=("dump_mock", "exception"),
    argvalues=test_write_parameters.values(),
    ids=test_write_parameters.keys(),
)
def test_write(
    dump_mock: mock.MagicMock, exception: Type[Exception], monkeypatch: MonkeyPatch
) -> None:
    """writeのテスト."""
    filename = "foo_file"

    open_mock = mock.MagicMock()
    monkeypatch.setattr(yaml, "dump", dump_mock)
    monkeypatch.setattr("builtins.open", open_mock)

    sut = LocalTestCaseRepositoryImpl(filename)

    if exception:
        with pytest.raises(exception):
            sut.write(testcases)
    else:
        sut.write(testcases)

    open_mock.assert_called_once_with(filename, "wt")
    dump_mock.assert_called_once_with([testcases[0].dict()], ANY, sort_keys=False)


test_read_parameters = {
    "OK": [mock.MagicMock(return_value=[testcases[0]]), None],
    "Error(読み込み失敗)": [mock.MagicMock(side_effect=OSError()), ReadError],
    "Error(Parse失敗)": [mock.MagicMock(return_value=["foo"]), ParseError],
}


@pytest.mark.parametrize(
    argnames=("safe_load_mock", "exception"),
    argvalues=test_read_parameters.values(),
    ids=test_read_parameters.keys(),
)
def test_read(
    safe_load_mock: mock.MagicMock, exception: Type[Exception], monkeypatch: MonkeyPatch
) -> None:
    """readのテスト."""
    filename = "foo_file"
    open_mock = mock.MagicMock()
    monkeypatch.setattr("builtins.open", open_mock)
    monkeypatch.setattr(yaml, "safe_load", safe_load_mock)

    sut = LocalTestCaseRepositoryImpl(filename)

    if exception:
        with pytest.raises(exception):
            sut.read()
    else:
        result = sut.read()
        assert result == testcases



"""テストケースの永続化を行う."""
from typing import List
from typing import Protocol

import yaml

from atcoder_helper.models.atcoder_test_case import AtcoderTestCase
from atcoder_helper.repositories.errors import ParseError
from atcoder_helper.repositories.errors import ReadError
from atcoder_helper.repositories.errors import WriteError


class LocalTestCaseRepository(Protocol):
    """テストケースの永続化を行うプロトコル."""

    def write(self, test_cases: List[AtcoderTestCase]) -> None:
        """書き込みを行う.

        Args:
            test_cases (List[TestCase]): 取得したテストスイート

        Raises:
            WriteError: 書き込み失敗
        """

    def read(self) -> List[AtcoderTestCase]:
        """読み込みを行う.

        Returns:
            List[TestCase]: 読み込まれたテストスイート

        Raises:
            ReadError: データの読み込みに失敗した
            ParseError: パースに失敗した
        """


def get_default_local_test_case_repository() -> LocalTestCaseRepository:
    """TestCaseRepositoryの標準実装を返す."""
    default_testcase_file = "testcases.yaml"
    return LocalTestCaseRepositoryImpl(default_testcase_file)


class LocalTestCaseRepositoryImpl:
    """テストケースの永続化を行う."""

    def __init__(self, filename: str):
        """__init__.

        Args:
            filename (str): 永続化先のファイル名
        """
        self._filename = filename

    def write(self, test_cases: List[AtcoderTestCase]) -> None:
        """書き込みを行う.

        Args:
            test_cases (List[TestCase]): 取得したテストスイート

        Raises:
            WriteError: 書き込み失敗
        """

        def str_representer(dumper: yaml.dumper.Dumper, data: str) -> yaml.Node:
            if len(data.splitlines()) > 1:  # check for multiline string
                return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
            return dumper.represent_scalar("tag:yaml.org,2002:str", data)

        yaml.add_representer(str, str_representer)

        try:
            with open(self._filename, "wt") as file:
                yaml.dump([case.dict() for case in test_cases], file, sort_keys=False)
        except OSError as e:
            raise WriteError(f"cannot open {self._filename}") from e

    def read(self) -> List[AtcoderTestCase]:
        """読み込みを行う.

        Returns:
            List[TestCase]: 読み込まれたテストスイート

        Raises:
            ReadError: データの読み込みに失敗した
            ParseError: パースに失敗した
        """
        try:
            with open(self._filename, "rt") as file:
                objects = yaml.safe_load(file)
        except OSError as e:
            raise ReadError(f"cannot open {self._filename}") from e

        try:
            return [AtcoderTestCase.parse_obj(object) for object in objects]
        except Exception as e:
            raise ParseError(
                f"failed to parse {self._filename} as AtcoderTestCase"
            ) from e



"""テストケースにまつわるデータ構造を定義する."""
from dataclasses import dataclass
from enum import Enum
from typing import Optional
from typing import cast

from colorama import Fore
from colorama import Style
from pydantic import BaseModel


class AtCoderTestStatus(Enum):
    """テスト結果の状態を保持する."""

    AC = 1
    WA = 2
    ERROR = 3
    JUSTSHOW = 4

    def _dye(self, message: str, color: str) -> str:
        return cast(str, color + message + Style.RESET_ALL)

    @property
    def dyed(self) -> str:
        """テスト結果の状態をふさわしい色に染めた文字列を返す."""
        table = {
            AtCoderTestStatus.AC: Fore.GREEN,
            AtCoderTestStatus.WA: Fore.YELLOW,
            AtCoderTestStatus.ERROR: Fore.RED,
            AtCoderTestStatus.JUSTSHOW: Fore.BLUE,
        }

        return self._dye(self.name, table[self])


@dataclass
class AtCoderTestResult:
    """入力と出力からなる一度のテストに対して、テスト実行にまつわる結果を保持する."""

    name: str
    status: AtCoderTestStatus
    actual: str
    error: str
    expected: Optional[str] = None


class AtcoderTestCase(BaseModel):
    """テストケースを表す."""

    name: str
    given: str
    expected: Optional[str]



"""atcoder_test_case_repoのテスト."""

from typing import Type

import mock
import pytest
import requests
from pytest import MonkeyPatch

from atcoder_helper.repositories.atcoder_test_case_repo import (
    AtCoderTestCaseRepositoryImpl,
)
from atcoder_helper.repositories.errors import ParseError

fetch_test_cases_parameters = {
    "OK": [
        mock.MagicMock(
            return_value=mock.MagicMock(
                text='<div id="task-statement"><span class="lang-ja"></span></div>'
            )
        ),
        None,
    ],
    "Error(通信エラー)": [
        mock.MagicMock(side_effect=Exception()),
        ConnectionError,
    ],
    "Error(parse error)": [
        mock.MagicMock(return_value=mock.MagicMock(text="<foo />")),
        ParseError,
    ],
}


@pytest.mark.parametrize(
    argnames=("get_mock", "exception"),
    argvalues=fetch_test_cases_parameters.values(),
    ids=fetch_test_cases_parameters.keys(),
)
def test_fetch_test_cases(
    get_mock: mock.MagicMock, exception: Type[Exception], monkeypatch: MonkeyPatch
) -> None:
    """fetch_test_casesのテスト."""
    contest = "foo_contest"
    task = "bar_task"
    session = requests.Session()

    sut = AtCoderTestCaseRepositoryImpl()

    monkeypatch.setattr(requests.Session, "get", get_mock)

    if exception:
        with pytest.raises(exception):
            sut.fetch_test_cases(session, contest, task)
    else:
        sut.fetch_test_cases(session, contest, task)



"""Taskディレクトリを初期化するためのservice."""
from typing import Optional
from typing import Protocol

import atcoder_helper.repositories.errors as repo_errors
from atcoder_helper.models.atcoder_task_config import TaskConfig
from atcoder_helper.repositories.atcoder_helper_config_repo import ConfigRepository
from atcoder_helper.repositories.atcoder_helper_config_repo import (
    get_default_config_repository,
)
from atcoder_helper.repositories.errors import ParseError
from atcoder_helper.repositories.errors import ReadError
from atcoder_helper.repositories.task_config_repo import TaskConfigRepository
from atcoder_helper.repositories.task_config_repo import (
    get_default_task_config_repository,
)
from atcoder_helper.services.errors import ConfigAccessError


class InitTaskDirService(Protocol):
    """TaskDirectoryを初期化するサービスのプロトコル."""

    def init_task(
        self,
        dir: Optional[str] = None,
        contest: Optional[str] = None,
        task: Optional[str] = None,
    ) -> None:
        """taskディレクトリを初期化します.

        Raises:
            DirectoryNotEmpty: 作成しようとしているディレクトリが空でない
            ConfigAccessError: 設定ファイルの読み書きに失敗
        """


def get_default_init_task_dir_service() -> InitTaskDirService:
    """InitTaskDirServiceの標準実装を返す.

    Returns:
        InitTaskDirService: _description_
    """
    return InitTaskDirServiceImpl(
        atcoder_helper_config_repo=get_default_config_repository(),
        task_config_repo=get_default_task_config_repository(),
    )


class InitTaskDirServiceImpl:
    """TaskDirectoryを初期化するサービス."""

    _atcoder_helper_config_repo: ConfigRepository
    _task_config_repo: TaskConfigRepository

    def __init__(
        self,
        atcoder_helper_config_repo: ConfigRepository,
        task_config_repo: TaskConfigRepository,
    ):
        """__init.

        Args:
            atcoder_helper_config_repo (ConfigRepository): _
            task_config_repo (TaskConfigRepository): _
        """
        self._atcoder_helper_config_repo = atcoder_helper_config_repo
        self._task_config_repo = task_config_repo

    def init_task(
        self,
        dir: Optional[str] = None,
        contest: Optional[str] = None,
        task: Optional[str] = None,
    ) -> None:
        """taskディレクトリを初期化します.

        Raises:
            DirectoryNotEmpty: 作成しようとしているディレクトリが空でない
            ConfigAccessError: 設定ファイルの読み書きに失敗
        """
        try:
            language_config = (
                self._atcoder_helper_config_repo.read().default_language_config
            )
        except (ReadError, ParseError) as e:
            raise ConfigAccessError("全体設定ファイルの読み込みに失敗しました") from e

        task_config = TaskConfig(
            build=language_config.build,
            run=language_config.run,
            contest=contest,
            task=task,
        )

        try:
            self._task_config_repo.write(
                task_config=task_config,
                template_dir=language_config.resolved_template_dir,
                target_dir=dir,
            )
        except repo_errors.DirectoryNotEmpty as e:
            raise ConfigAccessError("ディレクトリを初期化できません") from e
        except repo_errors.WriteError as e:
            raise ConfigAccessError("タスク設定ファイルの書き込みに失敗しました") from e
        except repo_errors.CopyError as e:
            raise ConfigAccessError("テンプレートによる初期化に失敗しました") from e



"""デフォルト言語を変更する."""


import os
from typing import Dict
from typing import Protocol

import atcoder_helper
from atcoder_helper.models.atcoder_helper_config import LanguageConfig
from atcoder_helper.repositories import errors as repository_errors
from atcoder_helper.repositories.atcoder_helper_config_repo import ConfigRepository
from atcoder_helper.repositories.atcoder_helper_config_repo import ConfigRepositoryImpl
from atcoder_helper.repositories.atcoder_helper_config_repo import (
    get_default_config_repository,
)
from atcoder_helper.services.errors import ConfigAccessError
from atcoder_helper.services.errors import UndefinedLanguage


class AtCoderHelperConfigService(Protocol):
    """AtCoderHelperConfigを操作する Serviceのプロトコル."""

    def config_use(self, language: str) -> None:
        """デフォルト言語を変更する.

        Raises:
            UndefinedLanguage: 存在しない言語をデフォルトに設定しようとした
            ConfigAccessError: 通信層のエラー
        """

    def config_default_language(self) -> LanguageConfig:
        """デフォルト言語を取得する.

        Raises:
            ConfigAccerssError: 設定ファイル読み書きのエラー

        Returns:
            LanguageConfig: デフォルト言語
        """

    def config_languages(self) -> Dict[str, LanguageConfig]:
        """使用可能な言語の一覧を取得する.

        Raises:
            ConfigAccessError: 設定ファイルの読み書きの失敗

        Returns:
            Dict[str, LanguageConfig]: 言語名から言語設定をひく辞書
        """

    def init_config(
        self,
    ) -> None:
        """configを初期化するためのservice.

        Raises:
            ConfigAccessError:
        """


def get_default_atcoder_helper_config_service() -> AtCoderHelperConfigService:
    """AtCoderHelperConfigServiceの標準実装を返す.

    Returns:
        AtCoderHelperConfigService:
    """
    return AtCoderHelperConfigServiceImpl(
        config_repo=get_default_config_repository(),
        default_config_repo=ConfigRepositoryImpl(
            os.path.join(
                atcoder_helper.__path__[0], "default_configs", "default_config.yaml"
            )
        ),
    )


class AtCoderHelperConfigServiceImpl:
    """AtCoderHelperConfigを操作する Service."""

    _config_repo: ConfigRepository
    _default_config_repo: ConfigRepository

    def __init__(
        self, config_repo: ConfigRepository, default_config_repo: ConfigRepository
    ):
        """_init_.

        Args:
            config_repo (ConfigRepository, optional): _
            default_config_repo (ConfigRepository, optional): _
        """
        self._config_repo = config_repo
        self._default_config_repo = default_config_repo

    def config_use(self, language: str) -> None:
        """デフォルト言語を変更する.

        Raises:
            UndefinedLanguage: 存在しない言語をデフォルトに設定しようとした
            ConfigAccessError: 通信層のエラー
        """
        try:
            config = self._config_repo.read()
            if language not in config.languages:
                raise UndefinedLanguage(f"言語{language}は設定の中に存在しません。")

            config.default_language = language
            self._config_repo.write(config)
        except (
            repository_errors.WriteError,
            repository_errors.ReadError,
            repository_errors.ParseError,
        ) as e:
            raise ConfigAccessError("設定ファイルの読み書きに失敗しました。") from e

    def config_default_language(self) -> LanguageConfig:
        """デフォルト言語を取得する.

        Raises:
            ConfigAccerssError: 設定ファイル読み書きのエラー

        Returns:
            LanguageConfig: デフォルト言語
        """
        try:
            config = self._config_repo.read()
        except (repository_errors.ReadError, repository_errors.ParseError) as e:
            raise ConfigAccessError("設定ファイルの読み込みに失敗しました") from e

        return config.languages[config.default_language]

    def config_languages(self) -> Dict[str, LanguageConfig]:
        """使用可能な言語の一覧を取得する.

        Raises:
            ConfigAccessError: 設定ファイルの読み書きの失敗

        Returns:
            Dict[str, LanguageConfig]: 言語名から言語設定をひく辞書
        """
        try:
            config = self._config_repo.read()
        except (repository_errors.ReadError, repository_errors.ParseError) as e:
            raise ConfigAccessError("設定ファイルの読み込みに失敗しました") from e

        return config.languages

    def init_config(
        self,
    ) -> None:
        """configを初期化するためのservice.

        Raises:
            ConfigAccessError:
        """
        try:
            default_atcoder_config = self._default_config_repo.read()
        except (repository_errors.ReadError, repository_errors.ParseError) as e:
            raise ConfigAccessError("デフォルト設定ファイルの読み込みに失敗しました.") from e

        try:
            self._config_repo.write(default_atcoder_config)
        except repository_errors.WriteError as e:
            raise ConfigAccessError("設定ファイルの書き込みに失敗しました.") from e



"""AtCoder用のコマンド.

以下の構成からなる
- atcoder_helper.scripts: 実行のentry_pointを定義する
- atcoder_helper.services: まとまったひとつのユースケース群を定義する
- atcoder_helper.models: 主要なデータ構造を定義する
- atcoder_helper.repositories: 通信層を定義する

"""



"""loginしているかどうかを取得するrepository."""

from typing import Protocol

import requests

from atcoder_helper.repositories.errors import ReadError
from atcoder_helper.repositories.utils import AtCoderURLProvider


class LoginStatusRepo(Protocol):
    """login情報を取得するrepositoryのプロトコル."""

    def is_logged_in(self, session: requests.Session) -> bool:
        """loginしているかどうかを判定する.

        Raises:
            ReadError: タスクGETに失敗

        Returns:
            bool: loginしているか否か
        """


def get_default_login_status_repo() -> LoginStatusRepo:
    """LoginStatusRepoの標準実装を返す.

    Returns:
        LoginStatusRepo: 標準実装
    """
    return LoginStatusRepoImpl()


class LoginStatusRepoImpl:
    """login情報を取得するrepository."""

    _url_provider = AtCoderURLProvider

    def is_logged_in(self, session: requests.Session) -> bool:
        """loginしているかどうかを判定する.

        Raises:
            ReadError: タスクGETに失敗

        Returns:
            bool: loginしているか否か
        """
        # たたもさんの atcoder-cli を参考にしている
        #  https://github.com/Tatamo/atcoder-cli/blob/0ca0d088f28783a4804ad90d89fc56eb7ddd6ef4/src/atcoder.ts#L46

        try:
            res = session.get(
                self._url_provider.submit_url("abc001"),
                allow_redirects=False,
            )
        except Exception as e:
            raise ReadError(
                f"cannot GET {self._url_provider.submit_url('abc001')}"
            ) from e

        return res.status_code == 200  # login していなければ302 redirect になる



"""Tests for atcoder_helper_config_repository."""


import os
from typing import Type

import mock
import pytest
import yaml
from mock import ANY
from pytest import MonkeyPatch

from atcoder_helper.models.atcoder_helper_config import AtCoderHelperConfig
from atcoder_helper.models.atcoder_helper_config import LanguageConfig
from atcoder_helper.repositories.atcoder_helper_config_repo import ConfigRepositoryImpl
from atcoder_helper.repositories.errors import ParseError
from atcoder_helper.repositories.errors import ReadError
from atcoder_helper.repositories.errors import WriteError

helper_config = AtCoderHelperConfig(
    languages={"foo": LanguageConfig(name="foo", build=[], run=[])},
    default_language="foo",
)

helper_config_dict = {
    "languages": [{"name": "foo", "build": [], "run": []}],
    "default_language": "foo",
}


test_config_repository_impl_read_parameters = {
    "OK": [mock.MagicMock(return_value=helper_config_dict), None, helper_config],
    "Error(読み込み失敗)": [mock.MagicMock(side_effect=OSError), ReadError, ANY],
    "Error(パース失敗)": [mock.MagicMock(return_value={}), ParseError, ANY],
}


@pytest.mark.parametrize(
    argnames=("safe_load_mock", "exception", "result"),
    argvalues=test_config_repository_impl_read_parameters.values(),
    ids=test_config_repository_impl_read_parameters.keys(),
)
def test_config_repository_impl_read(
    safe_load_mock: mock.MagicMock,
    exception: Type[Exception],
    result: AtCoderHelperConfig,
    monkeypatch: MonkeyPatch,
) -> None:
    """readのテスト."""
    filename = "file_foo"

    monkeypatch.setattr("builtins.open", mock.MagicMock())
    monkeypatch.setattr(yaml, "safe_load", safe_load_mock)

    sut = ConfigRepositoryImpl(filename=filename)

    if exception:
        with pytest.raises(exception):
            sut.read()
    else:
        result = sut.read()
        assert result == helper_config


@pytest.mark.parametrize(
    argnames=("dump_mock", "exception"),
    argvalues=[
        [mock.MagicMock(), None],
        [mock.MagicMock(side_effect=OSError()), WriteError],
    ],
    ids=["OK", "dumpのエラーでエラー"],
)
def test_config_repository_impl_write(
    dump_mock: mock.MagicMock, exception: Type[Exception], monkeypatch: MonkeyPatch
) -> None:
    """writeのテスト."""
    filename = "file_foo"
    sut = ConfigRepositoryImpl(filename=filename)

    open_mock = mock.MagicMock()

    monkeypatch.setattr(os, "makedirs", mock.MagicMock())
    monkeypatch.setattr("builtins.open", open_mock)
    monkeypatch.setattr(yaml, "dump", dump_mock)

    if exception:
        with pytest.raises(exception):
            sut.write(helper_config)
    else:
        sut.write(helper_config)

    open_mock.assert_called_once_with(filename, "wt")
    dump_mock.assert_called_once_with(helper_config.dict(), ANY)



"""atcoder_helper_configのテスト."""

import os
from typing import Optional

import pytest

import atcoder_helper
from atcoder_helper.models.atcoder_helper_config import LanguageConfig


@pytest.mark.parametrize(
    argnames=("conf", "template_dir"),
    argvalues=[
        [LanguageConfig(name="foo", build=[], run=[]), None],
        [
            LanguageConfig(name="foo", template_dir="bar", build=[], run=[]),
            "bar",
        ],
        [
            LanguageConfig(name="foo", use_default_template=True, build=[], run=[]),
            os.path.join(
                atcoder_helper.__path__[0], "default_configs", "templates", "foo"
            ),
        ],
    ],
    ids=["指定なし", "指定あり", "default templateを使う"],
)
def test_language_config_resolved_template_dir(
    conf: LanguageConfig, template_dir: Optional[str]
) -> None:
    """template_dirのテスト."""
    assert conf.resolved_template_dir == template_dir


@pytest.mark.skip(reason="自明なので後で書く")
def test_atcoder_helper_config_dict() -> None:
    """dict()のテスト."""
    pass



"""Services共通のutil."""

import os
from typing import Final

default_atcoder_helper_config_file: Final[str] = os.path.join(
    os.path.expanduser("~"), ".atcoder_helper", "config.yaml"
)


def get_atcoder_helper_config_filepath() -> str:
    """atcoder_helper_configのパスを決定する."""
    filepath = os.environ.get("ATCODER_HELPER_CONFIG_FILEPATH")
    if filepath:
        return filepath
    else:
        return default_atcoder_helper_config_file



"""AtcoderHelperConfigを定義する."""

import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from pydantic import BaseModel

import atcoder_helper


class LanguageConfig(BaseModel):
    """言語ごとの設定を保持する."""

    name: str
    template_dir: Optional[str]
    use_default_template: Optional[bool]
    build: List[str]
    run: List[str]

    @property
    def resolved_template_dir(self) -> Optional[str]:
        """実際に使用するtemplate_directory."""
        if self.use_default_template:
            return os.path.join(
                atcoder_helper.__path__[0],
                "default_configs",
                "templates",
                self.name,
            )
        else:
            return self.template_dir


class AtCoderHelperConfig(BaseModel):
    """atcoder_helper アプリ全体の設定を保持する."""

    languages: dict[str, LanguageConfig]
    default_language: str

    @property
    def default_language_config(self) -> LanguageConfig:
        """デフォルトの言語設定."""
        return self.languages[self.default_language]

    def dict(self, **_: Any) -> Dict[str, Any]:
        """辞書に変換."""
        return {
            "languages": [
                language_config.dict(exclude_none=True)
                for language_config in self.languages.values()
            ],
            "default_language": self.default_language,
        }



"""Description of this package."""
import setuptools

from atcoder_helper._version import __version__

with open("README.md", "r") as readme:
    long_description = readme.read()

AUTHOR = "yuchiki"
EMAIL = "yuki.imai77@gmail.com"
URL = "https://github.com/yuchiki/atcoder_helper"


setuptools.setup(
    name="atcoder_helper",
    author=AUTHOR,
    author_email=EMAIL,
    package_data={"atcoder_helper": ["default_configs/*"]},
    maintainer=AUTHOR,
    maintainer_email=EMAIL,
    include_package_data=True,
    description="automation cli tools for AtCoder",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT LICENSE",
    version=__version__,
    url=URL,
    download_url=URL,
    python_requires=">=3.10",
    install_requires=["colorama", "beautifulsoup4", "requests", "pyyaml", "pydantic"],
    entry_points={"console_scripts": "atcoder_helper=atcoder_helper.scripts.main:main"},
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""repositories.errorsの定義."""


class ReadError(Exception):
    """読み込めなかったエラー."""


class WriteError(Exception):
    """書き込めなかったエラー."""


class ParseError(Exception):
    """パースに失敗したエラー."""


class AlreadyLoggedIn(Exception):
    """既にログインしているエラー."""


class DirectoryNotEmpty(Exception):
    """Directoryが空でないエラー."""


class CopyError(Exception):
    """Copy時のエラー."""


class LoginFailure(Exception):
    """loginに失敗した."""


class ConnectionError(Exception):
    """通信に失敗した."""



"""TaskConfigを取得する."""
import os
import shutil
from typing import Optional
from typing import Protocol

import yaml

from atcoder_helper.models.atcoder_task_config import TaskConfig
from atcoder_helper.repositories.errors import CopyError
from atcoder_helper.repositories.errors import DirectoryNotEmpty
from atcoder_helper.repositories.errors import ParseError
from atcoder_helper.repositories.errors import ReadError
from atcoder_helper.repositories.errors import WriteError


class TaskConfigRepository(Protocol):
    """TaskConfigを取得するプロトコル.

    TaskConfig用ファイルは、Taskディレクトリにおかれていることを想定している.
    """

    def read(self) -> TaskConfig:
        """読み込みを行う.

        Returns:
            TaskConfig: 読み込んだTaskConfig

        Raises:
            ReadError: 読み込みに失敗した
            ParseError: パースに失敗した
        """

    def write(
        self,
        task_config: TaskConfig,
        template_dir: Optional[str] = None,
        target_dir: Optional[str] = None,
    ) -> None:
        """TaskConfigを書き込む.

        Args:
            task_config (TaskConfig):
            template_dir (Optional[str]): Defaults to None
            target_dir (Optional[str]): Defaults to None (current directory)


        Raises:
            DirectoryNotEmpty: ディレクトリが空でない
            WriteError: タスク設定ファイルの書き込みに失敗
            CopyError: テンプレートのコピーに失敗
        """


def get_default_task_config_repository() -> TaskConfigRepository:
    """TaskConfigRepositoryの標準実装."""
    default_filename = ".atcoder_helper_task_config.yaml"
    return TaskConfigRepositoryImpl(default_filename)


class TaskConfigRepositoryImpl:
    """TaskConfigを取得する.

    TaskConfig用ファイルは、Taskディレクトリにおかれていることを想定している.
    """

    _filename: str

    def __init__(self, filename: str):
        """__init__.

        Args:
            filename (str): TaskConfigがおかれているファイル名
        """
        self._filename = filename

    def read(self) -> TaskConfig:
        """読み込みを行う.

        Returns:
            TaskConfig: 読み込んだTaskConfig

        Raises:
            ReadError: 読み込みに失敗した
            ParseError: パースに失敗した
        """
        try:
            with open(self._filename) as file:
                object = yaml.safe_load(file)
        except OSError as e:
            raise ReadError(f"cannot read from {file}") from e

        try:
            return TaskConfig.parse_obj(object)
        except Exception as e:
            raise ParseError(f"{self._filename} can not read as TaskConfig") from e

    def write(
        self,
        task_config: TaskConfig,
        template_dir: Optional[str] = None,
        target_dir: Optional[str] = None,
    ) -> None:
        """TaskConfigを書き込む.

        Args:
            task_config (TaskConfig):
            template_dir (Optional[str]): Defaults to None
            target_dir (Optional[str]): Defaults to None (current directory)

        Raises:
            DirectoryNotEmpty: ディレクトリが空でない
            WriteError: タスク設定ファイルの書き込みに失敗
            CopyError: テンプレートのコピーに失敗
        """
        if target_dir is None:
            task_dir = os.path.dirname(os.path.abspath(self._filename))
        else:
            task_dir = target_dir

        def _is_empty(dir: str) -> bool:
            return len(os.listdir(dir)) == 0

        os.makedirs(task_dir, exist_ok=True)
        if not _is_empty(task_dir):
            raise DirectoryNotEmpty(f"directory {task_dir} is not empty")

        if template_dir is not None:
            try:
                print(f"{template_dir} -> {task_dir}")
                shutil.copytree(template_dir, task_dir, dirs_exist_ok=True)
            except OSError as e:
                raise CopyError("テンプレートディレクトリのコピー中にエラーが発生しました") from e

        try:
            with open(os.path.join(task_dir, self._filename), "wt") as file:
                yaml.dump(
                    task_config.dict(exclude_none=True),
                    file,
                    sort_keys=False,
                )
        except OSError as e:
            raise WriteError("タスク設定ファイルの初期化中にエラーが発生しました") from e



"""service層のエラーたち."""


class AlreadyLoggedIn(Exception):
    """既にログインしている."""


class ConfigAccessError(Exception):
    """設定ファイル読み書きの起因のエラー."""


class AtcoderAccessError(Exception):
    """atcoderページとの通信起因のエラー."""


class UndefinedLanguage(Exception):
    """未定義の言語."""


class DirectoryNotEmpty(Exception):
    """Directoryが空でないエラー."""
